/*
 * Copyright (C) 2013-2014,2017-2018 Christian Pierre MOMON
 * 
 * This file is part of Devinsy-xml.
 * 
 * Devinsy-xml is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Devinsy-xml is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Devinsy-xml.  If not, see <http://www.gnu.org/licenses/>
 */
package fr.devinsy.xml;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.zip.Deflater;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import org.apache.commons.lang3.StringUtils;

/**
 * The Class XMLZipWriter.
 * 
 * @author Christian Pierre MOMON (christian.momon@devinsy.fr)
 */
public class XMLZipWriter extends XMLWriter
{
	private ZipOutputStream stream;

	/**
	 * Instantiates a new XML zip writer.
	 * 
	 * @param file
	 *            the file
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	public XMLZipWriter(final File file) throws IOException
	{
		super();

		this.stream = new ZipOutputStream(new FileOutputStream(file));
		this.stream.setLevel(Deflater.BEST_COMPRESSION);
		this.stream.setMethod(ZipOutputStream.DEFLATED);
		this.stream.setComment("Generated by XMLZipWriter");
		this.stream.putNextEntry(new ZipEntry(buildEntryName(file)));
		this.out = new PrintWriter(new OutputStreamWriter(this.stream, "UTF-8"));
	}

	/**
	 * Instantiates a new XML zip writer.
	 * 
	 * @param file
	 *            the file
	 * @param generator
	 *            the generator
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	public XMLZipWriter(final File file, final String generator) throws IOException
	{
		super();

		this.stream = new ZipOutputStream(new FileOutputStream(file));
		this.stream.setLevel(Deflater.BEST_COMPRESSION);
		this.stream.setMethod(ZipOutputStream.DEFLATED);
		this.stream.setComment(generator);
		this.stream.putNextEntry(new ZipEntry(buildEntryName(file)));
		this.out = new PrintWriter(new OutputStreamWriter(this.stream, "UTF-8"));
	}

	/**
	 * Instantiates a new XML zip writer.
	 * 
	 * @param target
	 *            the target
	 * @param generator
	 *            the generator
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	public XMLZipWriter(final OutputStream target, final String generator) throws IOException
	{
		super();

		this.stream = new ZipOutputStream(target);
		this.stream.setLevel(Deflater.BEST_COMPRESSION);
		this.stream.setMethod(ZipOutputStream.DEFLATED);
		if (generator != null)
		{
			this.stream.setComment(generator);
		}
		this.out = new PrintWriter(new OutputStreamWriter(this.stream, "UTF-8"));
	}

	/**
	 * Instantiates a new XML zip writer.
	 * 
	 * @param target
	 *            the target
	 * @param fileName
	 *            the file name
	 * @param generator
	 *            the generator
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	public XMLZipWriter(final OutputStream target, final String fileName, final String generator) throws IOException
	{
		super();

		this.stream = new ZipOutputStream(target);
		this.stream.setLevel(Deflater.BEST_COMPRESSION);
		this.stream.setMethod(ZipOutputStream.DEFLATED);
		if (StringUtils.isNotBlank(generator))
		{
			this.stream.setComment(generator);
		}
		openEntry(fileName);
		this.out = new PrintWriter(new OutputStreamWriter(this.stream, "UTF-8"));
	}

	/**
	 * Close.
	 * 
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	@Override
	public void close() throws IOException
	{
		closeEntry();
		super.close();
	}

	/**
	 * Close entry.
	 * 
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	public void closeEntry() throws IOException
	{
		flush();
		this.stream.closeEntry();
	}

	/**
	 * Open entry.
	 * 
	 * @param fileName
	 *            the file name
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	public void openEntry(final String fileName) throws IOException
	{
		if (StringUtils.isBlank(fileName))
		{
			throw new IllegalArgumentException("fileName is blank.");
		}
		else
		{
			this.stream.putNextEntry(new ZipEntry(buildEntryName(new File(fileName))));
		}
	}

	/**
	 * Builds the entry name.
	 * 
	 * @param source
	 *            the source
	 * @return the string
	 */
	private static String buildEntryName(final File source)
	{
		String result;

		if (source == null)
		{
			throw new IllegalArgumentException("Null parameter.");
		}
		else
		{
			result = source.getName();

			int separatorIndex = result.lastIndexOf('.');
			if (separatorIndex > 0)
			{
				result = result.substring(0, separatorIndex);
			}

			if (!result.endsWith(".xml"))
			{
				result += ".xml";
			}
		}

		//
		return result;
	}
}
